
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: C_FriendList, C_PartyInfo, ChatFrame_SendBNetTell, ChatFrame_SendTell, IsInGroup, LE_PARTY_CATEGORY_HOME, QuickJoinRoleSelectionFrame, SetItemRef, UnitIsGroupAssistant, UnitIsGroupLeader

local CHARACTER_FRIEND = CHARACTER_FRIEND
local CreateFromMixins = CreateFromMixins
local FAVORITES = FAVORITES
local FRIENDS = FRIENDS
local getmetatable = getmetatable
local GetNormalizedRealmName = GetNormalizedRealmName
local Mixin = Mixin
local next = next
local setmetatable = setmetatable
local string = string
local table = table
local type = type
local UNKNOWN = UNKNOWN

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local MAJOR, MINOR = "SixSocial-1.0", 1
if not LibStub then error(MAJOR .. " requires LibStub.") end
local L = LibStub:NewLibrary(MAJOR, MINOR)
if not L then return end

local AceEvent = LibStub:GetLibrary("AceEvent-3.0", true)
if not AceEvent then return end
AceEvent:Embed(L)

local AceTimer = LibStub:GetLibrary("AceTimer-3.0", true)
if not AceTimer then return end
AceTimer:Embed(L)

L.MATCH_RICH_PRESENCE = {"^(.-) %- ([%a' ]+)$", "^(.-) %- ([%a' ]+) %((.-)%)$"}
L.PLAYER_FACTION_GROUP = UnitFactionGroup("player")
L.REALM_NAME = GetNormalizedRealmName()

function L.SendInvite(target)
	local groupGUID = L:GetGroupGuid(target.accountName, target.guid, target.unit)

	if groupGUID and not IsInGroup(LE_PARTY_CATEGORY_HOME) then
		QuickJoinRoleSelectionFrame:ShowForGroup(groupGUID)
	elseif target.unit then
		if not IsInGroup() or UnitIsGroupLeader("player") or UnitIsGroupAssistant("player") then
			C_PartyInfo.InviteUnit(target.unit)
		else
			C_PartyInfo.RequestInviteFromUnit(target.unit)
		end
	end
end

function L.SendTell(target)
	if target.accountName then
		ChatFrame_SendBNetTell(target.accountName)
	elseif target.accountID then
		SetItemRef("BNplayer:" .. target.name .. ":" .. target.accountID, target.name, "LeftButton")
	elseif target.unit then
		ChatFrame_SendTell(target.unit)
	end
end

function L.SendWho(target)
	if target.unit then
		C_FriendList.SendWho("x-" .. target.unit)
	end
end

-------------------------------------------------------------------------------
-- Enumerators
-------------------------------------------------------------------------------

L.Enum = {}

L.Enum.ClientName = {
	RTRO = "Blizzard Arcade Collection",
	VIPR = "Call of Duty: Black Ops 4",
	ZEUS = "Call of Duty: Black Ops Cold War",
	ODIN = "Call of Duty: MW",
	LAZR = "Call of Duty: MW2CR",
	FORE = "Call of Duty: Vanguard",
	WLBY = "Crash Bandicoot 4",
	DST2 = "Destiny 2",
	OSI = "Diablo II: Resurrected",
	D3 = "Diablo III",
	Fen = "Diablo IV",
	ANBS = "Diablo Immortal",
	WTCG = "Hearthstone",
	Hero = "Heroes of the Storm",
	Pro = "Overwatch",
	S2 = "StarCraft II",
	S1 = "StarCraft: Remastered",
	GRY = "Warcraft Arclight Rumble",
	W3 = "Warcraft III: Reforged",
	WoW = "World of Warcraft",
}

L.Enum.Club = {
	Broadcast = "broadcast",
	Delete = "Delete",
	Id = "id",
	Initialize = "Initialize",
	IterateMembers = "IterateMembers",
	MemberCount = "memberCount",
	Name = "name",
	Type = "type",
	Update = "Update",
}

L.Enum.ClubType = Enum.ClubType

L.Enum.ClubMember = {
	AccountId = "accountID",
	ClassName = "className",
	ClubId = "clubID",
	ClubType = "clubType",
	Delete = "Delete",
	Guid = "guid",
	Id = "id",
	Initialize = "Initialize",
	Level = "level",
	MythicRating = "mythicRating",
	Name = "name",
	Note = "note",
	Rank = "rank",
	RankOrder = "rankOrder",
	SendInvite = "SendInvite",
	SendTell = "SendTell",
	SendWho = "SendWho",
	Status = "status",
	StatusOrder = "statusOrder",
	Unit = "unit",
	Update = "Update",
	Zone = "zone",
}

L.Enum.Friend = {
	AccountId = "accountID",
	AccountName = "accountName",
	Activity = "activity",
	BattleTag = "battleTag",
	Broadcast = "broadcast",
	Category = "category",
	CharacterName = "characterName",
	ClassName = "className",
	ClientName = "clientName",
	ClientTexture = "clientTexture",
	FactionName = "factionName",
	FriendGuid = "friendGUID",
	Index = "Index",
	Initialize = "Initialize",
	IsActive = "isActive",
	IsFavorite = "isFavorite",
	Level = "level",
	Name = "name",
	Note = "note",
	SendInvite = "SendInvite",
	SendTell = "SendTell",
	SendWho = "SendWho",
	Status = "status",
	StatusOrder = "statusOrder",
	Type = "type",
	Unit = "unit",
	Update = "Update",
}

L.Enum.FriendType = {
	Any = 0,
	BattleNet = 1,
	Character = 2,
	Player = 3,
	None = 4,
	Unknown = 5,
}

L.Enum.Status = {
	Away = 0,
	Busy = 1,
	Mobile = 2,
	Offline = 3,
	Online = 4,
}

L.Enum.StatusOrder = {
	[L.Enum.Status.Away] = 2,
	[L.Enum.Status.Busy] = 1,
	[L.Enum.Status.Mobile] = 2,
	[L.Enum.Status.Offline] = 3,
	[L.Enum.Status.Online] = 1,
}

L.Enum.StatusText = {
	[L.Enum.Status.Away] = FRIENDS_LIST_AWAY,
	[L.Enum.Status.Busy] = FRIENDS_LIST_BUSY,
	[L.Enum.Status.Mobile] = "Mobile",
	[L.Enum.Status.Offline] = FRIENDS_LIST_OFFLINE,
	[L.Enum.Status.Online] = FRIENDS_LIST_ONLINE,
}

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------

L.CLUB = setmetatable({
	[L.Enum.Club.Broadcast] = nil,
	[L.Enum.Club.Delete] = nil,
	[L.Enum.Club.Id] = nil,
	[L.Enum.Club.Initialize] = nil,
	[L.Enum.Club.IterateMembers] = nil,
	[L.Enum.Club.MemberCount] = 0,
	[L.Enum.Club.Name] = nil,
	[L.Enum.Club.Type] = nil,
	[L.Enum.Club.Update] = nil,
}, {
	__eq = function (a, b)
		local Key = L.Enum.Club

		return a[Key.Id] == b[Key.Id]
	end,
	__lt = function (a, b)
		local Key = L.Enum.Club
		local ZZZ = "ZZZ"

		return (a[Key.Name] or ZZZ) < (b[Key.Name] or ZZZ)
	end,
})

L.CLUB_MEMBER = setmetatable({
	[L.Enum.ClubMember.AccountId] = nil,
	[L.Enum.ClubMember.ClassName] = nil,
	[L.Enum.ClubMember.ClubId] = nil,
	[L.Enum.ClubMember.ClubType] = nil,
	[L.Enum.ClubMember.Delete] = nil,
	[L.Enum.ClubMember.Guid] = nil,
	[L.Enum.ClubMember.Id] = nil,
	[L.Enum.ClubMember.Initialize] = nil,
	[L.Enum.ClubMember.Level] = 0,
	[L.Enum.ClubMember.MythicRating] = 0,
	[L.Enum.ClubMember.Name] = nil,
	[L.Enum.ClubMember.Note] = nil,
	[L.Enum.ClubMember.Rank] = nil,
	[L.Enum.ClubMember.RankOrder] = 0,
	[L.Enum.ClubMember.SendInvite] = L.SendInvite,
	[L.Enum.ClubMember.SendTell] = L.SendTell,
	[L.Enum.ClubMember.SendWho] = L.SendWho,
	[L.Enum.ClubMember.Status] = L.Enum.Status.Offline,
	[L.Enum.ClubMember.StatusOrder] = 0,
	[L.Enum.ClubMember.Unit] = nil,
	[L.Enum.ClubMember.Update] = nil,
	[L.Enum.ClubMember.Zone] = nil,
}, {
	__eq = function (a, b)
		local Key = L.Enum.ClubMember

		return a[Key.Id] == b[Key.Id]
	end,
	__lt = function (a, b)
		local Key = L.Enum.ClubMember
		local ZZZ = "ZZZ"

		if a[Key.StatusOrder] ~= b[Key.StatusOrder] then
			return a[Key.StatusOrder] < b[Key.StatusOrder]
		elseif a[Key.Level] ~= b[Key.Level] then
			return a[Key.Level] > b[Key.Level]
		elseif a[Key.RankOrder] ~= b[Key.RankOrder] then
			return a[Key.RankOrder] < b[Key.RankOrder]
		else
			return (a[Key.Name] or ZZZ):upper() < (b[Key.Name] or ZZZ):upper()
		end
	end,
})

L.CLUB_MEMBER_RANK = {
	[_G.Enum.ClubRoleIdentifier.Owner] = COMMUNITY_MEMBER_ROLE_NAME_OWNER,
	[_G.Enum.ClubRoleIdentifier.Leader] = COMMUNITY_MEMBER_ROLE_NAME_LEADER,
	[_G.Enum.ClubRoleIdentifier.Moderator] = COMMUNITY_MEMBER_ROLE_NAME_MODERATOR,
	[_G.Enum.ClubRoleIdentifier.Member] = COMMUNITY_MEMBER_ROLE_NAME_MEMBER,
}

L.CLUB_MEMBER_STATUS = {
	[_G.Enum.ClubMemberPresence.Online] = L.Enum.Status.Online,
	[_G.Enum.ClubMemberPresence.OnlineMobile] = L.Enum.Status.Mobile,
	[_G.Enum.ClubMemberPresence.Away] = L.Enum.Status.Away,
	[_G.Enum.ClubMemberPresence.Busy] = L.Enum.Status.Busy,
	[_G.Enum.ClubMemberPresence.Offline] = L.Enum.Status.Offline,
	[_G.Enum.ClubMemberPresence.Unknown] = L.Enum.Status.Offline,
}

L.FRIEND = setmetatable({
	[L.Enum.Friend.AccountId] = nil,
	[L.Enum.Friend.AccountName] = nil,
	[L.Enum.Friend.Activity] = nil,
	[L.Enum.Friend.BattleTag] = nil,
	[L.Enum.Friend.Broadcast] = nil,
	[L.Enum.Friend.Category] = nil,
	[L.Enum.Friend.CharacterName] = nil,
	[L.Enum.Friend.ClassName] = nil,
	[L.Enum.Friend.ClientName] = nil,
	[L.Enum.Friend.ClientTexture] = nil,
	[L.Enum.Friend.FactionName] = nil,
	[L.Enum.Friend.FriendGuid] = nil,
	[L.Enum.Friend.Index] = nil,
	[L.Enum.Friend.Initialize] = nil,
	[L.Enum.Friend.IsActive] = false,
	[L.Enum.Friend.IsFavorite] = false,
	[L.Enum.Friend.Level] = nil,
	[L.Enum.Friend.Name] = nil,
	[L.Enum.Friend.Note] = nil,
	[L.Enum.Friend.SendInvite] = L.SendInvite,
	[L.Enum.Friend.SendTell] = L.SendTell,
	[L.Enum.Friend.SendWho] = L.SendWho,
	[L.Enum.Friend.Status] = L.Enum.Status.Offline,
	[L.Enum.Friend.StatusOrder] = 0,
	[L.Enum.Friend.Type] = L.Enum.FriendType.Unknown,
	[L.Enum.Friend.Unit] = nil,
	[L.Enum.Friend.Update] = nil,
}, {
	__eq = function (a, b)
		local Key = L.Enum.Friend

		return (a[Key.AccountId] or a[Key.FriendGuid]) == (b[Key.AccountId] or b[Key.FriendGuid])
	end,
	__lt = function (a, b)
		local Key = L.Enum.Friend
		local ZZZ = "ZZZ"

		if a[Key.BattleTag] ~= b[Key.BattleTag] then
			return (a[Key.BattleTag] or ZZZ):upper() < (b[Key.BattleTag] or ZZZ):upper()
		else
			return (a[Key.CharacterName] or ZZZ):upper() < (b[Key.CharacterName] or ZZZ):upper()
		end
	end,
})

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

function L.falsy(var)
	return var and var ~= "" and var ~= 0 and var or nil
end

function L.FormatUnit(name, realm, richPresence)
	if not L.REALM_NAME then
		L.REALM_NAME = GetNormalizedRealmName()
	end

	if name then
		if name:match("-") then
			name, realm = string.split("-", name)
		end

		if not L.falsy(realm) then
			realm = (not richPresence and L.REALM_NAME) or L.GetRealmFromRichPresence(richPresence) or UNKNOWN
		end

		return realm == L.REALM_NAME and name or name .. "-" .. realm
	end
end

function L.GetFriendCategory(type, note, isFavorite)
	return note and note:match("%((.-)%)")
		or type == L.Enum.FriendType.Character and CHARACTER_FRIEND
		or isFavorite and FAVORITES
		or FRIENDS
end

function L.GetFriendNote(note)
	return note and L.falsy(note:gsub(" ?%((.-)%) ?", ""))
end

function L.GetRealmFromRichPresence(richPresence)
	for _, pattern in next, L.MATCH_RICH_PRESENCE do
		if richPresence:match(pattern) then
			return richPresence:gsub(pattern, "%2"):gsub(" ", "")
		end
	end
end

function L.list(iterator)
	local result = {}

	for item in iterator do
		table.insert(result, item)
	end

	return result
end

function L.merge(...)
	local mt = {}
	for _, t in next, {...} do Mixin(mt, getmetatable(t) or {}) end
	return setmetatable(CreateFromMixins(...), mt)
end

function L.upper(var)
	return type(var) == "string" and var:upper() or var
end

-------------------------------------------------------------------------------
-- Library
-------------------------------------------------------------------------------

L.timer = {}

function L:SendMessageWithDelay(...)
	local key = table.concat({...}, ":")
	if self.timer[key] then self:CancelTimer(self.timer[key]) end
	self.timer[key] = self:ScheduleTimer("SendMessage", 0.25, ...)
end
